﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using STFXmlParser;
using PdfDocument;
using CalcHelper;

namespace Selections
{
    // Used when sending checkbox selections to Checkbox Control.
    public enum cbcSelection { selCtrlBoxSet, selNonCtrlBox, unSelCtrlBox };

    public enum boxSelTypes { allSel, leftToolBeltOpWaits, leftSel, leftSelDict, rightCtrlSel, rightCalcSel };
    public enum boxSelOpLock { none, div, mult, sub, add };

    /*
     * A custom EventArgs class to pass box unselection info
     * to the Tool Belt's Box Selector.
     */
    public class BoxUnSelArgs : EventArgs
    {
        public string msg { get; set; }
        public List<string> boxes { get; set; }
    }

    public class BoxSelections
    {
        /*
         * ***********************
         * This area is used when we send checkbox selections to
         * Checbox Control:
         * 
         * This is a clone of the CBC 'Check Box Sets'. It
         * has commas but no '[' and no ']'.
         */
        public List<string> cbcBoxSetsList = new List<string>();

        public int cbcCurIndex = -1;

        /*
         * The page number of the current check box set. We do
         * not allow the user to select checkboxes on different
         * pages to be in the same set.
         */
        public int cbcCurPage;

        public cbcSelection cbcSelections;

        /*
         * ***********************
         */ 

        // true = New box format (box23) else it is the
        // old box format (box12_33).
        private bool _newBoxFormat;

        /*
         * Indicates if any selections should be sent on to
         * Vscode. It may turn out that what looks like a
         * selection via a click or lasso turns out to be
         * an unselection, ex: a selected box is clicked 
         * which is an unselection, so we just unselect and
         * that is that.
         */
        public bool toVscode;

        // Create the unselection event itself:
        public static event EventHandler<BoxUnSelArgs> BoxUnSelEvent;

        /*
         * ToolBelt Box Selector box formats.
         * 
         * NOTE1: These need to follow the same order as the ToolBelt's
         * Box Selector boxFormatComboBox's list elements.
         * 
         * NOTE2: boxSelFormatsIndex has a default value of 0 at Calc Helper
         * startup as per below.
         */
        public List<string> boxSelFormats = new List<string> { "{0}.to_f()", "{0}.to_i()", "{0}", "\"{0}\"", "{0}.to_f().abs()" };
        public int boxSelFormatsIndex = 0;

        public boxSelOpLock bsOpLockState;

        // Might not use this, maybe just the enum itself?
        public boxSelTypes boxSelType;

        // Click [no lasso] only:
        public RightCtrlBoxSel rightCtrlBoxSel;
        public RightCalcBoxSel rightCalcBoxSel;

        // Click or lasso:
        public LeftBoxSel leftBoxSel;
        public Dictionary<string, LeftBoxSel> leftBoxSelDict;

        /*
         * These are currently selected boxes that are to be unselected.
         * It is possible that some of these may be toolBeltOpWaitBoxes
         * in which case we will need to tell the Tool Belt's Box
         * Selector to handle them.
         */
        public List<string> unselLeftBoxes;

        /*
         * These are the ToolBelt's operator wait boxes. Store here and in
         * the Tool Belts text box while we wait for the user to select an
         * operator to be used with the boxes, like '+', then we will send
         * the box add string [using the '+' example] on to the associated
         * Vscode file.
         * 
         * These boxes may also be used for an individually boxed calc
         * rather than a mathematical operation.
         */
        public List<string> toolBeltOpWaitBoxes;

        public BoxSelections()
        {
            toVscode = false;
            rightCtrlBoxSel = null;
            rightCalcBoxSel = null;
            leftBoxSelDict = new Dictionary<string, LeftBoxSel>();
            toolBeltOpWaitBoxes = new List<string>();
            unselLeftBoxes = new List<string>();
            bsOpLockState = boxSelOpLock.none;
        }

        // Trigger the unselection event:
        protected virtual void OnBoxUnSelEvent(BoxUnSelArgs e)
        {
            if (BoxUnSelEvent != null)
            {
                BoxUnSelEvent(this, e);
            }
        }

        /*
         * Called From:
         * 
         * SetupPdf1() - Methods.cs.
         *  
         * newBoxFormat: Tells us the box format being used in the document's
         *               code - true = New format (box12), false = old format
         *               (box1_12).
         */
        public void BoxSelSetBoxFormat(bool newBoxFormat)
        {
            _newBoxFormat = newBoxFormat;
        }

        /*
         * Called From:
         * 
         *  toolBelt_ToolBeltEvent() - Methods2.cs. ToolBelt event handler.
         *  
         *  ToolBeltClosed() - Form1.cs.
         *  refreshBtn_Click() - After refreshing Calc Helper with info
         *                       from a document's Javascript file.
         *
         *  ProcessBoxSelMode() - Methods2.cs.
         *  ProcessLeftMouseClick() - Methods2.cs
         *  ProcessLeftMouseBoxNumMagClick() - Methods2.cs
         *  ProcessLeftMouseUp() - Methods2.cs
         * 
         * We check the selType arg to see which box selections to unselect.
         * Could be all, could be one. We abort this function if we ever fail
         * to get a field array list when we should get one - should not happen.
         * 
         * selType: This is the type of selection we are to reset (unselect).
         *      
         * pa1: This is the PageArray for pdf1. We need this to reset the box fields
         *      on pages that contain selections.
         *      
         * pdf1: This is for displaying a pdf page, the main display panel.
         * 
         * pa2: This is the PageArray for pdf2. We will make sure this is not null,
         *      this would be if the screen was not split. If it is split then we
         *      need to make the same control updates to the same fields.
         *      
         * pdf2: This is for displaying a pdf page, the second display panel provided
         *         we have a split screen, else this will be null.
         */
        public void ResetSelections(boxSelTypes selType, ref PageArray pa1, ref Pdf pdf1, ref PageArray pa2, ref Pdf pdf2)
        {
            // Caller has ensured that pa1 and pdf1 are not null.

            int curPage;
            int paPage;
            bool split = false;
            List<Field> fa1 = null;
            List<Field> fa2 = null;
            
            /*
             * We will need to check if pa2 and pdf2 are null as
             * we may not have a split screen. If the screen is split
             * we will need to unselect the same fields in pa2.
             */
            if ((pdf2 != null) && (pa2 != null))
            {
                split = true;
            }

            try
            {
                if (selType == boxSelTypes.allSel)
                {
                    paPage = -1;

                    foreach (KeyValuePair<string, LeftBoxSel> kvp in leftBoxSelDict)
                    {
                        curPage = pa1.GetBoxPageNum(kvp.Key);

                        SetRefreshPdfPage(ref pdf1, ref pdf2, split, curPage);

                        if (curPage != paPage)
                        {
                            paPage = curPage;

                            pa1.GetFieldArrayRef(curPage - 1, ref fa1);

                            if (fa1 == null)
                            {
                                NullFieldArrayError("ResetSelections", "pa1", kvp.Key);
                                return;
                            }

                            if (split == true)
                            {
                                pa2.GetFieldArrayRef(curPage - 1, ref fa2);

                                if (fa2 == null)
                                {
                                    NullFieldArrayError("ResetSelections", "pa2", kvp.Key);
                                    return;
                                }
                            }
                        }

                        fa1[kvp.Value.faIndex].leftBoxSel = false;

                        if (split == true)
                        {
                            /*
                             * By definition the pa's and fa's are identical since
                             * they are both tied to the same document.
                            */
                            fa2[kvp.Value.faIndex].leftBoxSel = false;
                        }
                    }

                    // Now clear the dictionary:
                    leftBoxSelDict.Clear();

                    /* IF toolBeltOpWaitBoxes is not empty:
                     * 
                     * Inform the Tool Belt that it needs to empty it's operator wait
                     * text box 'holding area' since we are unselecting all boxes in
                     * the left mouse box selection dictionary. The operator wait
                     * boxes are a subset of the boxes in the dictionary. The Tool
                     * Belt's Box Selector will also clear the associated list
                     * toolBeltOpWaitBoxes.
                     * 
                     * IF the Tool Belt Box Selector triggered this unselection of
                     * boxes it would be via boxSelTypes.allSel in which case the
                     * Box Selector would have already cleared it's operator wait
                     * textbox holding area AND the toolBeltOpWaitBoxes list. There
                     * are several other callers of this function however, that can
                     * also trigger via boxSelTypes.allSel in which case we would
                     * need to send a message to the Tool Belt if we have any
                     * operator wait boxes in the list.
                     */
                    if (toolBeltOpWaitBoxes.Count > 0)
                    {
                        BoxUnSelArgs busa = new BoxUnSelArgs();

                        busa.boxes = new List<string>();
                        busa.boxes.AddRange(toolBeltOpWaitBoxes);

                        OnBoxUnSelEvent(busa);
                    }
                }

                if ((selType == boxSelTypes.allSel) || (selType == boxSelTypes.rightCalcSel))
                {
                    if (rightCalcBoxSel != null)
                    {
                        curPage = pa1.GetBoxPageNum(rightCalcBoxSel.box);

                        SetRefreshPdfPage(ref pdf1, ref pdf2, split, curPage);

                        pa1.GetFieldArrayRef(curPage - 1, ref fa1);

                        if (fa1 == null)
                        {
                            NullFieldArrayError("ResetSelections", "pa1", rightCalcBoxSel.box);
                            return;
                        }

                        if (split == true)
                        {
                            pa2.GetFieldArrayRef(curPage - 1, ref fa2);

                            if (fa2 == null)
                            {
                                NullFieldArrayError("ResetSelections", "pa2", rightCalcBoxSel.box);
                                return;
                            }
                        }

                        fa1[rightCalcBoxSel.faIndex].rightCalcBoxSel = false;

                        if (split == true)
                        {
                            fa2[rightCalcBoxSel.faIndex].rightCalcBoxSel = false;
                        }

                        rightCalcBoxSel = null;
                    }
                }

                if ((selType == boxSelTypes.allSel) || (selType == boxSelTypes.rightCtrlSel))
                {
                    if (rightCtrlBoxSel != null)
                    {
                        curPage = pa1.GetBoxPageNum(rightCtrlBoxSel.box);

                        SetRefreshPdfPage(ref pdf1, ref pdf2, split, curPage);

                        pa1.GetFieldArrayRef(curPage - 1, ref fa1);

                        if (fa1 == null)
                        {
                            NullFieldArrayError("ResetSelections", "pa1", rightCtrlBoxSel.box);
                            return;
                        }

                        if (split == true)
                        {
                            pa2.GetFieldArrayRef(curPage - 1, ref fa2);

                            if (fa2 == null)
                            {
                                NullFieldArrayError("ResetSelections", "pa2", rightCtrlBoxSel.box);
                                return;
                            }
                        }

                        fa1[rightCtrlBoxSel.faIndex].rightCtrlBoxSel = false;

                        if (split == true)
                        {
                            fa2[rightCtrlBoxSel.faIndex].rightCtrlBoxSel = false;
                        }

                        rightCtrlBoxSel = null;
                    }
                }

                /*
                 * The toolBeltSelAbort list boxes are 'inside' leftBoxSelDict.
                 * Only the Tool Belt itself [Box Selector] will trigger this
                 * via the selection abort button [the operator wait box
                 * selections] OR when the Tool Belt form closes, the event
                 * handler for that triggers this.
                 * 
                 * For the abort case the Tool Belt Box Selector will only
                 * clear it's operator wait textbox, we need the contents
                 * of the toolBeltOpWaitBoxes list to key into the left box
                 * selection dictionary, so we will clear the
                 * toolBeltOpWaitBoxes list here when finished with it.
                 */
                if ((selType == boxSelTypes.leftToolBeltOpWaits) && (toolBeltOpWaitBoxes.Count > 0))
                {
                    Dictionary<string, int> boxPageDict = new Dictionary<string, int>();

                    /*
                     * Sorting just on the boxes [new box format] is no guarantee you
                     * are sorting by page order which is the whole intent of sorting here.
                     * The old box format was sorting by page order, not anymore going
                     * forward. We are trying to avoid needless switching between field
                     * array fetching below.
                     * 
                     * The ONLY box on the document that we do not keep track of is
                     * box1_65535, so we can just ask pa1 for the page number here.
                     */
                    foreach (string box in toolBeltOpWaitBoxes)
                    {
                        if (!boxPageDict.ContainsKey(box))
                        {
                            boxPageDict.Add(box, pa1.GetBoxPageNum(box));
                        }
                    }

                    paPage = -1;

                    foreach (KeyValuePair<string, int> kvp in boxPageDict.OrderBy(key => key.Value))
                    {
                        curPage = kvp.Value;

                        SetRefreshPdfPage(ref pdf1, ref pdf2, split, curPage);

                        if (curPage != paPage)
                        {
                            paPage = curPage;
                            pa1.GetFieldArrayRef(curPage - 1, ref fa1);

                            if (fa1 == null)
                            {
                                NullFieldArrayError("ResetSelections", "pa1", kvp.Key);
                                return;
                            }

                            if (split == true)
                            {
                                pa2.GetFieldArrayRef(curPage - 1, ref fa2);

                                if (fa2 == null)
                                {
                                    NullFieldArrayError("ResetSelections", "pa2", kvp.Key);
                                    return;
                                }
                            }
                        }

                        // The box SHOULD be in this dictionary.
                        if (leftBoxSelDict.TryGetValue(kvp.Key, out leftBoxSel) == true)
                        {
                            fa1[leftBoxSel.faIndex].leftBoxSel = false;

                            if (split == true)
                            {
                                fa2[leftBoxSel.faIndex].leftBoxSel = false;
                            }

                            // Remove this box from dictionary:
                            leftBoxSelDict.Remove(kvp.Key);
                        }
                    }

                    // Empty out the toolBeltSelAbort list.
                    toolBeltOpWaitBoxes.Clear();
                }

                /*
                 * One or more left mouse clicked boxes are to be unselected.
                 * 
                 * We need to see if any of these boxes are boxes that the
                 * Tool Belt's Box Selector is waiting on the user for an
                 * operator or separator to be used in sending these on to
                 * Vscode. If so we will need to put these in a separate list
                 * and inform the Tool Belt via an unselection event.
                 */
                if ((selType == boxSelTypes.leftSel) && (unselLeftBoxes.Count > 0))
                {
                    BoxUnSelArgs busa = new BoxUnSelArgs();
                    busa.boxes = new List<string>();
                    Dictionary<string, int> boxPageDict = new Dictionary<string, int>();

                    /*
                     * Sorting just on the boxes [new box format] is no guarantee you
                     * are sorting by page order which is the whole intent of sorting here.
                     * The old box format was sorting by page order, not anymore going
                     * forward. We are trying to avoid needless switching between field
                     * array fetching below.
                     * 
                     * The ONLY box on the document that we do not keep track of is
                     * box1_65535, so we can just ask pa1 for the page number here.
                     */
                    foreach (string box in unselLeftBoxes)
                    {
                        if (!boxPageDict.ContainsKey(box))
                        {
                            boxPageDict.Add(box, pa1.GetBoxPageNum(box));
                        }
                    }

                    paPage = -1;

                    foreach (KeyValuePair<string, int> kvp in boxPageDict.OrderBy(key => key.Value))
                    {
                        curPage = kvp.Value;

                        SetRefreshPdfPage(ref pdf1, ref pdf2, split, curPage);

                        if (curPage != paPage)
                        {
                            paPage = curPage;
                            pa1.GetFieldArrayRef(curPage - 1, ref fa1);

                            if (fa1 == null)
                            {
                                NullFieldArrayError("ResetSelections", "pa1", kvp.Key);
                                return;
                            }

                            if (split == true)
                            {
                                pa2.GetFieldArrayRef(curPage - 1, ref fa2);

                                if (fa2 == null)
                                {
                                    NullFieldArrayError("ResetSelections", "pa2", kvp.Key);
                                    return;
                                }
                            }
                        }

                        // The box SHOULD be in this dictionary.
                        if (leftBoxSelDict.TryGetValue(kvp.Key, out leftBoxSel) == true)
                        {
                            fa1[leftBoxSel.faIndex].leftBoxSel = false;

                            if (split == true)
                            {
                                fa2[leftBoxSel.faIndex].leftBoxSel = false;
                            }

                            // Remove this box from dictionary:
                            leftBoxSelDict.Remove(kvp.Key);
                        }

                        if (toolBeltOpWaitBoxes.Contains(kvp.Key))
                        {
                            busa.boxes.Add(kvp.Key);
                        }
                    }

                    // Empty out the unselLeftBoxes list.
                    unselLeftBoxes.Clear();

                    // Fire event if any boxes:
                    if (busa.boxes.Count > 0)
                    {
                        OnBoxUnSelEvent(busa);
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), "ResetSelections():", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /*
         * This routine displays an error message to the user
         * if a null field array (List<Field>) was retrieved from
         * a PageArray pa1/pa2. This error should not happen, but
         * I want to know about it.
         * 
         * Called From:
         *      ResetSelections()
         *      ProcessRightCtrlBoxSel()
         *      ProcessRightCalcBoxSel()
         *      ProcessLeftBoxSel()
         *      
         * Arguments:
         *      fcn - The function that retrieved the null field array.
         *      arg1 - Used in string.Formatting the error message.
         *      arg2 - Used in string.Formatting the error message.
         */
        public void NullFieldArrayError(string fcn, string arg1, string arg2)
        {
            using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.error, string.Format("{0}() Error:\n\n{1}.GetFieldArrayRef() returned null field\narray while processing {2}.\n\nAborting {0}().", fcn, arg1, arg2)))
            {
                mainMsg.ShowDialog();
            }
        }

        /*
         * This checks to see if we are to set pdf1's and/or pdf2's
         * 'refreshPage' field during a selection reset. What this means
         * is that we need to visually update a pdf page if we unselect
         * a box on it, i.e. remove the selection rectangle(s).
         * 
         * Called From:
         * 
         *      ResetSelections()
         *      
         * Arguments:
         * 
         *      pdf1 - The main panel which we are checking to see if the
         *              current page is to be refreshed.
         *      pdf2 - Same reasoning as for pdf1, however this only applies
         *              when we have a split screen.
         *      split - 'true' means we have a split screen.
         *      curPage - The current page as relates to the box we are unselecting.
         *      
         * Callers have ensured that pdf1 is not null [which it should not be] and
         * if split is 'true' pdf2 is not null.
         */
        public void SetRefreshPdfPage(ref Pdf pdf1, ref Pdf pdf2, bool split, int curPage)
        {
            if (pdf1.refreshPage == false)
            {
                if (pdf1.curPage == curPage)
                {
                    pdf1.refreshPage = true;
                }
            }

            if ((split == true) && (pdf2.refreshPage == false))
            {
                if (pdf2.curPage == curPage)
                {
                    pdf2.refreshPage = true;
                }
            }
        }

        /*
         * This handles the selection of one or more boxes via the left
         * mouse lasso process as well as the left mouse click process.
         * Up to the user to do 'sane' lassoing.
         * 
         * We do not 'police' the user's lassoing here, they can select
         * already controlled boxes [perfectly normal] or uncontrolled
         * boxes, even select the box they are calculating which does
         * happen though not too often.
         *
         * We add these selections to the leftBoxSelDict.
         * 
         * Called From:
         * 
         *      ProcessLeftMouseUp()
         *      ProcessLeftMouseClick()
         *      ProcessLeftMouseBoxNumMagClick()
         *      
         * Arguments:
         * 
         *      leftBoxSel - Holds the left box selection info.
         *      pdf1 - The 'main' panel that is always active.
         *      pa1 - The PageArray structure which supports pdf1.
         *      pdf2 - The split panel - maybe active if a split screen.
         *      pa2 - The split panel's PageArray structure.
         *      
         * Returns 'true' if no problems encountered.
         * 
         * It has been verified that pdf1 and pa1 are not null.
         */
        public bool ProcessLeftBoxSel(ref List<LeftBoxSel> leftBoxSel, ref Pdf pdf1, ref PageArray pa1, ref Pdf pdf2, ref PageArray pa2)
        {
            bool result = false;
            bool split = false;
            List<Field> fa1 = null;
            List<Field> fa2 = null;
            int page;

            /*
             * All will be on the same page:
             */
            page = leftBoxSel[0].paIndex + 1;

            if ((pdf2 != null) && (pa2 != null))
            {
                /*
                 * We have a split screen, so we need to do the same
                 * selection to the other fa's selection field.
                */
                split = true;
            }

            pa1.GetFieldArrayRef(page - 1, ref fa1);

            if (fa1 == null)
            {
                NullFieldArrayError("ProcessLeftBoxSel", "pa1", "left box selections");
                return result;
            }

            if (split == true)
            {
                pa2.GetFieldArrayRef(page - 1, ref fa2);

                if (fa2 == null)
                {
                    NullFieldArrayError("ProcessLeftBoxSel", "pa2", "left box selections");
                    return result;
                }
            }

            SetRefreshPdfPage(ref pdf1, ref pdf2, split, page);

            foreach (LeftBoxSel lbs in leftBoxSel)
            {
                fa1[lbs.faIndex].leftBoxSel = true;

                if (split == true)
                {
                    /*
                     * By definition the pa's and fa's are identical since
                     * they are both tied to the same document.
                    */
                    fa2[lbs.faIndex].leftBoxSel = true;
                }

                // According to Webster's....
                if (leftBoxSelDict.ContainsKey(lbs.box))
                {
                    // Debug Check: This box should not be in the dictionary.
                    string msg;

                    msg = string.Format("ProcessLeftBoxSel() Error:\n\n{0} is already present in leftBoxSelDict.", lbs.box);
                    using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.warning, msg))
                    {
                        mainMsg.ShowDialog();
                    }
                }
                else
                {
                    leftBoxSelDict.Add(lbs.box, lbs);
                }
            }

            result = true;

            return result;
        }

        /*
         * This handles the selection of a non-controlled box via the right
         * mouse button click for the purpose of creating a calculation of
         * some sort - basically an assignment to a box.
         * 
         * We do not 'police' the user's keystrokes here, when the user
         * initiates this click they are basically creating a box assignment
         * which could be a blue field control calc, or a 'regular' calc,
         * no dataflow.
         *
         * We first unselect all boxes in preparation for the user's selections
         * regarding this calculating box.
         * 
         * If we are selecting the box we will set toVscode to 'true' to signal
         * to the caller that it is ok to send this box on to Vscode.
         * 
         * Called From:
         * 
         *      ProcessRightMouseClick()
         *      
         * Arguments:
         * 
         *      calcBoxSel - Holds the info on the box we are calculating.
         *      pdf1 - The 'main' panel that is always active.
         *      pa1 - The PageArray structure which supports pdf1.
         *      pdf2 - The split panel - maybe active if a split screen.
         *      pa2 - The split panel's PageArray structure.
         *      
         * Returns 'true' if no problems encountered.
         * 
         * It has been verified that pdf1 and pa1 are not null.
         */
        public bool ProcessRightCalcBoxSel(RightCalcBoxSel calcBoxSel, ref Pdf pdf1, ref PageArray pa1, ref Pdf pdf2, ref PageArray pa2)
        {
            bool result = false;
            bool selectBox = true;
            bool split = false;
            List<Field> fa1 = null;
            List<Field> fa2 = null;
            int page;

            if ((rightCalcBoxSel != null) && (calcBoxSel.box == rightCalcBoxSel.box))
            {
                /*
                 * They just clicked the present box being calculated, so
                 * we treat this as an unselection, so we are done.
                 */
                selectBox = false;
                toVscode = false;
            }

            // We will clear the 'chalkboard' of all selections:
            ResetSelections(boxSelTypes.allSel, ref pa1, ref pdf1, ref pa2, ref pdf2);

            if (selectBox == true)
            {
                toVscode = true;
                page = calcBoxSel.paIndex + 1;

                if ((pdf2 != null) && (pa2 != null))
                {
                    /*
                     * We have a split screen, so we need to do the same
                     * selection to the other fa's selection field.
                    */
                    split = true;
                }

                SetRefreshPdfPage(ref pdf1, ref pdf2, split, page);

                rightCalcBoxSel = new RightCalcBoxSel(calcBoxSel.box, calcBoxSel.paIndex, calcBoxSel.faIndex);

                pa1.GetFieldArrayRef(page - 1, ref fa1);

                if (fa1 == null)
                {
                    NullFieldArrayError("ProcessRightCalcBoxSel", "pa1", calcBoxSel.box);
                    return result;
                }

                if (split == true)
                {
                    pa2.GetFieldArrayRef(page - 1, ref fa2);

                    if (fa2 == null)
                    {
                        NullFieldArrayError("ProcessRightCalcBoxSel", "pa2", calcBoxSel.box);
                        return result;
                    }
                }

                fa1[calcBoxSel.faIndex].rightCalcBoxSel = true;

                if (split == true)
                {
                    /*
                     * By definition the pa's and fa's are identical since
                     * they are both tied to the same document.
                    */
                    fa2[calcBoxSel.faIndex].rightCalcBoxSel = true;
                }
            }

            result = true;

            return result;
        }

        /*
         * This handles the selection of a controlled box via the right
         * mouse button click. So someone just right mouse clicked on a box
         * that is either dataflowed, calculated, checkbox controlled, or
         * bluefield controlled.
         * 
         * **********
         * 
         * If the box is currently selected:
         * 
         * If a left mouse selection - We will keep the selection indication
         * for the left mouse selection since the user has selected this box
         * for insertion into vscode and we want to keep the visual indication
         * of this selection.
         * 
         * If a right mouse selection - If the selected box is the same box
         * as the currently selected box in rightCtrlBoxSel we will unselect
         * the box and that will be the end of it. If they differ, then we 
         * unselect the current rightCtrlBoxSel, then select the 'new' current
         * control box and set rightCtrlBoxSel to it. 
         * 
         * **********
         * 
         * If the box is not currently selected:
         * 
         * We will first unselect the current rightCtrlBoxSel box if there
         * is one, then select this box.
         * 
         * Called From:
         * 
         *      ProcessRightMouseClick()
         *      
         * Arguments:
         * 
         *      ctrlBoxSel - Holds the info on the box just clicked on.
         *      pdf - The panel from which the click originated, either pdf1 or pdf2.
         *      pa - The PageArray structure which supports pdf, either pa1 or pa2.
         *      pdfSplit - The split panel's pdf - either pdf1 or pdf2 - opposite of pdf.
         *      paSplit - The split panel's PageArray structure, either pa1 or pa2 -
         *                  opposite of pa.
         *      docFile - The PdfDocument object.
         *      
         * Returns 'true' if we are to go to the control code in Vscode, 'false' if not.
         */
        public bool ProcessRightCtrlBoxSel(RightCtrlBoxSel ctrlBoxSel, ref Pdf pdf, ref PageArray pa, ref Pdf pdfSplit, ref PageArray paSplit, ref DocFile docFile)
        {
            bool goToCode = true;
            bool selectBox = true;
            int unselectPage = -1;
            int selectPage = -1;
            int curPage = (int)(pdf.curPage - 1);
            int splitCurPage = -1;
            List<Field> fa = null;
            List<Field> faSplit = null;

            if ((pdfSplit != null) && (paSplit != null))
            {
                /*
                 * We have a split screen, so we need to do the same
                 * changes to the other fa's selection fields concerning
                 * selecting and unselecting.
                 * 
                 * We also need to show those changes IF the splitscreen's
                 * pdf is displaying the same page for a selection or unselection.
                */
                splitCurPage = (int)(pdfSplit.curPage - 1);
            }

            /*
             * Now determine if the box is currently selected - check first for a 
             * left mouse button selection [we will still go to the control code
             * if the box is left mouse selected, we just will not alter the
             * current left selection indication]:
             */
            if (leftBoxSelDict.ContainsKey(ctrlBoxSel.box))
            {
                selectBox = false;
            }

            if (rightCtrlBoxSel != null)
            {
                // Unselect the rightCtrlBoxSel box:
                pa.GetFieldArrayRef(rightCtrlBoxSel.paIndex, ref fa);

                if (fa == null)
                {
                    NullFieldArrayError("ProcessRightCtrlBoxSel", "pa", rightCtrlBoxSel.box);
                    return false;
                }

                fa[rightCtrlBoxSel.faIndex].rightCtrlBoxSel = false;
                unselectPage = rightCtrlBoxSel.paIndex;

                // Do the same if we have a split screen:
                if (splitCurPage > -1)
                {
                    paSplit.GetFieldArrayRef(rightCtrlBoxSel.paIndex, ref faSplit);

                    if (faSplit == null)
                    {
                        NullFieldArrayError("ProcessRightCtrlBoxSel", "pa", rightCtrlBoxSel.box);
                        return false;
                    }

                    faSplit[rightCtrlBoxSel.faIndex].rightCtrlBoxSel = false;
                }

                if (selectBox == true)
                {
                    if (rightCtrlBoxSel.box != ctrlBoxSel.box)
                    {
                        rightCtrlBoxSel.box = ctrlBoxSel.box;
                        rightCtrlBoxSel.paIndex = ctrlBoxSel.paIndex;
                        rightCtrlBoxSel.faIndex = ctrlBoxSel.faIndex;
                    }
                    else
                    {
                        selectBox = false;
                        rightCtrlBoxSel = null;
                        goToCode = false;
                    }
                }
                else
                {
                    rightCtrlBoxSel = null;
                }
            }
            else
            {
                if (selectBox == true)
                {
                    rightCtrlBoxSel = new RightCtrlBoxSel(ctrlBoxSel.box, ctrlBoxSel.paIndex, ctrlBoxSel.faIndex, ctrlBoxSel.ctrlType);
                }
            }

            if (selectBox)
            {
                // Do the selection:

                pa.GetFieldArrayRef(ctrlBoxSel.paIndex, ref fa);

                if (fa == null)
                {
                    NullFieldArrayError("ProcessRightCtrlBoxSel", "pa", ctrlBoxSel.box);
                    return false;
                }

                fa[ctrlBoxSel.faIndex].rightCtrlBoxSel = true;
                selectPage = ctrlBoxSel.paIndex;

                // Do the same if we have a split screen:
                if (splitCurPage > -1)
                {
                    paSplit.GetFieldArrayRef(ctrlBoxSel.paIndex, ref faSplit);

                    if (faSplit == null)
                    {
                        NullFieldArrayError("ProcessRightCtrlBoxSel", "pa", ctrlBoxSel.box);
                        return false;
                    }

                    faSplit[ctrlBoxSel.faIndex].rightCtrlBoxSel = true;
                }
            }

            /*
             * Now we need to check if we show the current page to reflect
             * a selection and/or an unselection. This check is also done
             * if we have a splitscreen.
             */
            if ((unselectPage == curPage) || (selectPage == curPage))
            {
                pdf.centerPic = false;
                pdf.ShowPage(ref docFile.hPdf, (int)pdf.curPage, false, ref pa);
                pdf.centerPic = true;
            }

            if (splitCurPage > -1)
            {
                if ((unselectPage == splitCurPage) || (selectPage == splitCurPage))
                {
                    pdfSplit.centerPic = false;
                    pdfSplit.ShowPage(ref docFile.hPdf, (int)pdfSplit.curPage, false, ref paSplit);
                    pdfSplit.centerPic = true;
                }
            }

            return goToCode;
        }
    }

    /*
     * A right mouse button clicked
     * Dataflow|Calculated|Bluefield Control|Checkbox Control box.
     */
    public class RightCtrlBoxSel
    {
        public string box;
        public int paIndex;
        public int faIndex;
        public int ctrlType;

        public RightCtrlBoxSel(string boxSel, int paIndx, int faIndx, int ctrlId)
        {
            box = boxSel;
            paIndex = paIndx;
            faIndex = faIndx;
            ctrlType = ctrlId;
        }
    }

    /*
     * The box we are calculating, selected via
     * a right mouse button click.
     */
    public class RightCalcBoxSel
    {
        public string box;
        public int paIndex;
        public int faIndex;

        public RightCalcBoxSel(string boxSel, int paIndx, int faIndx)
        {
            box = boxSel;
            paIndex = paIndx;
            faIndex = faIndx;
        }
    }

    /*
     * A box selected via a left mouse button click/lasso.
     * 
     * Put the box name in here even though we have a dictionary
     * with the name as the key - gives more flexibility in
     * usage...at least that is the idea.
     */
    public class LeftBoxSel
    {
        public string box;
        public int paIndex;
        public int faIndex;

        public LeftBoxSel(string boxSel, int paIndx, int faIndx)
        {
            box = boxSel;
            paIndex = paIndx;
            faIndex = faIndx;
        }
    }

}